/*
 * Sistemas de Telecomunicações 
 *          2012/2013
 */
package protocol;

import simulator.Frame;
import terminal.NetworkLayer;

/**
 * Protocol 1 : Utopian protocol that does not retransmit failed frames
 * 
 * @author lflb@fct.unl.pt
 */
public class Protocol1 implements Callbacks {

    public Protocol1(Simulator _sim, NetworkLayer _net) {
        this.sim = _sim;
        this.net = _net;
        next_frame_to_send = decr_seq(0);
        frame_expected = 0;
    }

    /**
     * Increment one unit a sequence number wrapping when the maximum sequence 
     * number is reached
     * @param n sequence number
     * @return the successor of the sequence number
     */
    final int incr_seq(int n) {
        return (n + 1) % (sim.get_max_sequence() + 1);
    }

    /**
     * Decrement one unit a sequence number wrapping when the maximum sequence 
     * number is reached
     * @param n sequence number
     * @return the predecessor of the sequence number
     */
    final int decr_seq(int n) {
        return (sim.get_max_sequence() + n) % (sim.get_max_sequence() + 1);
    }

    /**
     * Fetches the network layer for the next packet and starts it transmission
     * @return true is started data frame transmission, false otherwise
     */
    private boolean send_next_data_packet() {
        String packet= net.from_network_layer();
        if (packet != null) {
            next_frame_to_send= incr_seq(next_frame_to_send);
            Frame frame = Frame.new_Data_Frame(next_frame_to_send, frame_expected, packet);
            sim.to_physical_layer(frame);
            return true;
        }
        return false;
    }
    
    /**
     * CALLBACK FUNCTION: handle the beginning of the simulation event
     * @param time current simulation time
     */
    @Override
    public void start_simulation(long time) {
        sim.Log("\nUtopian Protocol\n\n");
        send_next_data_packet();    // Start sending the first packet
    }

    /**
     * CALLBACK FUNCTION: handle the end of Data frame transmission, and start 
     * sending the next frame
     * @param time current simulation time
     * @param seq  sequence number of the Data frame transmitted
     */
    @Override
    public void handle_Data_end(long time, int seq) {
        send_next_data_packet();
    }

    /**
     * CALLBACK FUNCTION: handle the timer event; ignores it in this case
     * @param time current simulation time
     * @param key  timer key
     */
    @Override
    public void handle_Timer(long time, int key) {
        sim.Log(time + " Timeout " + key + "\n");
    }

    /**
     * CALLBACK FUNCTION: handle the ack timer event; ignores it in this case
     * @param time current simulation time
     */
    @Override
    public void handle_ack_Timer(long time) {
        sim.Log(time + " ACK Timeout\n");
    }
    
    /**
     * CALLBACK FUNCTION: handle the reception of a frame from the physical layer
     * @param time current simulation time
     * @param frame frame received
     */
    @Override
    public void from_physical_layer(long time, Frame frame) {
        sim.Log(time + " protocol1 received: " + frame.toString() +"\n");
        if (frame.kind() == Frame.DATA_FRAME) {     // Check the frame kind
            if (frame.seq() == frame_expected) {    // Check the sequence number
                net.to_network_layer(frame.info()); // Send the frame to the network layer
                frame_expected = incr_seq(frame_expected);
            }
        }
    }

    /**
     * CALLBACK FUNCTION: handle the end of the simulation
     * @param time current simulation time
     */
    @Override
    public void end_simulation(long time) {
        sim.Log("Stopping simulation\n");
    }
    
    
    /* Variables */
    
    /**
     * Reference to the simulator (Terminal), to get the configuration and send commands
     */
    private Simulator sim;
    
    /**
     * Reference to the network layer, to send a receive packets
     */
    private NetworkLayer net;
    
    /**
     * Sequence number of the next data frame
     */
    private int next_frame_to_send;
    
    /**
     * Expected sequence number of the next data frame received
     */
    private int frame_expected;
}
