/*
 * Sistemas de Telecomunicações 
 *          2012/2013
 */
package protocol;

import simulator.Frame;
import terminal.NetworkLayer;

/**
 * Protocol 4 : Go-back-N protocol with NACKs
 * 
 * @author ?????
 */
public class Protocol4 implements Callbacks {

    public Protocol4(Simulator _sim, NetworkLayer _net) {
        this.sim = _sim;
        this.net = _net;
        ack_expected = 0;
        buf_index = 0;
    }

    /**
     * Return true if a <= b < c circularly; false otherwise
     * @param a - lower bound
     * @param b - test value
     * @param c - upper bound
     * @return true is b is between a and c
     */
    static boolean between(int a, int b, int c) {
        return ((a<=b)&&(b<c)) || ((c<a)&&(a<=b)) || ((b<c)&&(c<a));
    }

    /**
     * Increment one unit a sequence number wrapping when the maximum sequence 
     * number is reached
     * @param n sequence number
     * @return the successor of the sequence number
     */
    final int incr_seq(int n) {
        return (n + 1) % (sim.get_max_sequence() + 1);
    }

    /**
     * Decrement one unit a sequence number wrapping when the maximum sequence 
     * number is reached
     * @param n sequence number
     * @return the predecessor of the sequence number
     */
    final int decr_seq(int n) {
        return (sim.get_max_sequence() + n) % (sim.get_max_sequence() + 1);
    }

    /**
     * Calculates the difference between two sequence numbers taking into 
     * account the wrapping when the maximum sequence number is reached
     * @param a first sequence number
     * @param b first sequence number
     * @return number of numbers between a and b
     */
    int diff_seq(int a, int b) {
        if (b>=a) {
            return b-a;
        }
        else {
            return b+sim.get_max_sequence()+1-a;
        }     
    }
    
    /**
     * Returns the index of the position a sequence number in the sending circular buffer
     * @param seq sequence number
     * @return the index of the sending buffer
     */
    int buf_index(int seq) {
        int n= (buf_index+diff_seq(ack_expected,seq)) % sim.get_send_window();
        return n;
    }
    
    /**
     * CALLBACK FUNCTION: handle the beginning of the simulation event
     * @param time current simulation time
     */
    @Override
    public void start_simulation(long time) {
        sim.Log("\nGo-Back-N Protocol with NACKs\n\n");
        sim.Log("\nProtocol not implemented yet\n");
    }

    /**
     * CALLBACK FUNCTION: handle the end of Data frame transmission, start timer
     * @param time current simulation time
     * @param seq  sequence number of the Data frame transmitted
     */
    @Override
    public void handle_Data_end(long time, int seq) {
        // Not implemented yet
    }
    
    /**
     * CALLBACK FUNCTION: handle the timer event; retransmit failed frames
     * @param time current simulation time
     * @param key  timer key
     */
    @Override
    public void handle_Timer(long time, int key) {
        // Not implemented yet
    }
    
    /**
     * CALLBACK FUNCTION: handle the ack timer event; send ACK frame
     * @param time current simulation time
     */
    @Override
    public void handle_ack_Timer(long time) {
        // Not implemented yet
    }

    /**
     * CALLBACK FUNCTION: handle the reception of a frame from the physical layer
     * @param time current simulation time
     * @param frame frame received
     */
    @Override
    public void from_physical_layer(long time, Frame frame) {
        // Not implemented yet
    }

    /**
     * CALLBACK FUNCTION: handle the end of the simulation
     * @param time current simulation time
     */
    @Override
    public void end_simulation(long time) {
        sim.Log("Stopping simulation\n");
    }
    
    
    /* Variables */
    
    /**
     * Reference to the simulator (Terminal), to get the configuration and send commands
     */
    private Simulator sim;
    
    /**
     * Reference to the network layer, to send a receive packets
     */
    private NetworkLayer net;
    
    /**
     * ACK sequence number expected (oldest sent data frame)
     */
    private int ack_expected;
    
    /**
     * Index of the ack_expected frame in the buffer
     */
    private int buf_index;
}
